"""Convert workflow processing."""

import asyncio
import logging
from pathlib import Path
from shutil import copy
from nicegui import ui
from .. import helpers

log = logging.getLogger(__name__)


@helpers.track_subprocess_cleanup
async def process_convert(output_widget, progress_widget, args, spectronaut_cmd, spectronaut_key):
    """Run Convert workflow"""
    if not args['datafiles']:
        ui.notify('No files to process', type='negative')
        return

    if args['output_directory'] == '':
        ui.notify('Output directory not specified', type='negative')
        return
    
    if not Path(args['output_directory']).exists():
        Path(args['output_directory']).mkdir(parents=True)

    data_folder = Path(args['output_directory']).joinpath('data')
    data_folder.mkdir(parents=True, exist_ok=True)
    params_folder = Path(args['output_directory']).joinpath('params')
    params_folder.mkdir(parents=True, exist_ok=True)

    # Make a snapshot of the parameters files in the output directory to ensure reproducibility
    for property_key in ['properties_file']:
        if args[property_key] != '' and Path(args[property_key]).exists():
            new_path = Path(params_folder).joinpath(Path(args[property_key]).name)
            copy(Path(args[property_key]), new_path)
            args[property_key] = new_path
    
    output_widget.clear()

    try:
        ui.notify('Preparing data files...', type='info')
        await helpers.prepare_datafiles_async(args['datafiles'], data_folder, log, progress_widget)
    except asyncio.CancelledError:
        log.warning('Processing cancelled by user')
        return
    except Exception as e:
        log.error(f'Error preparing data files: {e}')
        return
    
    args.pop('protocol')
    
    try:
        args_list = await asyncio.to_thread(helpers.get_args, args)
        log.debug(f'Got base arguments: {len(args_list)} included')
    except Exception as e:
        log.error(f'Cannot get arguments: {e}')
        return

    log.info('Activating Spectronaut')
    result = await helpers.run_cmd(spectronaut_cmd + ['activate', spectronaut_key], log)
    if result:
        log.info('Spectronaut activated successfully')
    else:
        log.error('Cannot activate Spectronaut, see detailed log')
        return
    
    global_args = helpers.get_global_args(args)
    if global_args:
        log.info('Executing global arguments')
        log.debug(f'Global arguments: {global_args}')
        result = await helpers.run_cmd(spectronaut_cmd + global_args, log)
        if result:
            log.info('Global arguments set successfully')
        else:
            log.error('Cannot execute global arguments, see detailed log')
            return

    total = len(args['datafiles'])
    log.info(f'Converting {total} files')
    progress_widget.visible = True
    progress_widget.value = 0
    success = True
    for i in range(total):
        file_path = Path(args['datafiles'][i]['path'])
        result = await helpers.run_cmd(spectronaut_cmd + ['convert', '-i', file_path] + args_list, log)
        success = success and result
        if result:
            log.info(f'[{i + 1}|{total}] Converted successfully')
        else:
            log.error('Processing failed, see detailed log')
        
        progress_widget.value = (i + 1) / total
    
    progress_widget.visible = False
    
    log.info('Deactivating Spectronaut')
    result = await helpers.run_cmd(spectronaut_cmd + ['deactivate'], log)
    if result:
        log.info('Spectronaut deactivated')
    else:
        log.error('Cannot deactivate Spectronaut, see detailed log')
    
    return success



import asyncio
import re
from nicegui import ui, events
from pathlib import Path
from typing import List, Set

class LocalPicker(ui.dialog):
    def __init__(self, directory: str = '~', *, 
                 multiple: bool = False, 
                 show_files: bool = True, 
                 on_select=None,
                 default_selection: str = ''):
        super().__init__()
        self.path = Path(directory).expanduser()
        self.multiple = multiple
        self.show_files = show_files
        self.on_select = on_select
        # default_selection: empty string means no filter
        self.current_filter = default_selection if default_selection else ''
        #multiple patterns
        if self.current_filter.find('|') != -1:
            pattern_str = '|'.join([f'{re.escape(p.strip())}' for p in self.current_filter.split('|')])
            self.regex_filter = re.compile(f'(.+)?({pattern_str})(.+)?', re.IGNORECASE)
        else:
            self.regex_filter = re.compile(f'(.+)?{re.escape(self.current_filter)}(.+)?', re.IGNORECASE)\
                  if self.current_filter else re.compile(r'.+?')
        
        # Track shift key state via a UI-level event
        self.shift_is_held = False
        
        self.selected: Set[str] = set() # Store as strings for JSON safety
        self.items: List[Path] = []
        self.last_clicked_index: int = None

        with self, ui.card().classes('w-[800px] h-[700px] column no-wrap'):
            # Detect Shift key globally while dialog is open
            ui.keyboard(on_key=self._handle_key)

            # Header: title, path display and controls
            ui.label('File/Folder Selector').classes('text-h6')
            ui.label('Hold SHIFT to select/unselect ranges').classes('text-caption text-blue')

            with ui.row().classes('w-full items-center gap-2'):
                # path_input lets the user type a path;
                self.path_input = ui.input(label='Path', value=str(self.path)).classes('grow text-xs text-grey-6 truncate')\
                    .on('keydown.enter', lambda e: self._set_path(self.path_input.value))
                ui.button('Create new', icon='create_new_folder', on_click=self._create_new_dir).props('icon')
                
            # Filter and show-files controls
            with ui.row().classes('items-center q-mt-sm gap-2'):
                ui.label('Filter:').classes('text-sm')
                # free-text filter where user can enter extension or pattern (e.g. .csv)
                self.filter_input = ui.input(value='' if self.current_filter == 'All Files' else self.current_filter,
                                             placeholder='any part of name').classes('w-1/3')\
                                                .on('keydown.enter', lambda e: self._set_filter(self.filter_input.value))
                self.show_files_checkbox = ui.checkbox('Show files', value=self.show_files,
                                                      on_change=lambda e: self._set_show_files(e.value))

            with ui.scroll_area().classes('col grow border rounded q-mt-sm'):
                self.list_container = ui.list().props('dense separtor')
                self.update_list()

            # Footer: selection count, quick actions and confirm/cancel
            with ui.row().classes('w-full justify-between mt-4 items-center'):
                ui.label().bind_text_from(self, 'selected', backward=lambda s: f'{len(s)} selected')
                with ui.row():
                    ui.button('Select All', on_click=self._select_all).props('flat')
                    ui.button('Clear', on_click=self._clear_selection).props('flat')
                    ui.button('Cancel', on_click=self.close).props('flat')
                    ui.button('Confirm', on_click=self._handle_confirm).props('primary')

    def _handle_key(self, e: events.KeyEventArguments):
        """Track shift key globally."""
        if e.key.shift:
            self.shift_is_held = e.action.keydown

    def _set_filter(self, value: str):
        """Set current filter (e.g. '.csv' or substring) and refresh list."""
        v = (value or '').strip()
        # empty string means no filter
        self.current_filter = v

        if self.current_filter.find('|') != -1:
            pattern_str = '|'.join([f'{re.escape(p.strip())}' for p in self.current_filter.split('|')])
            self.regex_filter = re.compile(f'(.+)?({pattern_str})(.+)?', re.IGNORECASE)
        else:
            self.regex_filter = re.compile(f'(.+)?{re.escape(self.current_filter)}(.+)?', re.IGNORECASE)\
                 if self.current_filter else re.compile(r'.+?')

        self.update_list()

    def _create_new_dir(self):
        """Ask for a directory name, create it inside the current path and navigate into it."""
        with ui.dialog() as dlg, ui.card().classes('q-pa-md'):
            ui.label('Create new directory').classes('text-lg q-mb-sm')
            name_input = ui.input(label='Directory name', placeholder='Enter new directory name').classes('w-full')

            def _on_create(*_):
                name = (name_input.value or '').strip()
                if not name:
                    ui.notify('Please enter a directory name', type='warning')
                    return
                new_path = self.path.joinpath(name)
                try:
                    new_path.mkdir(parents=False, exist_ok=False)
                except FileExistsError:
                    ui.notify('Directory already exists', type='negative')
                    return
                except Exception as e:
                    ui.notify(f'Error creating directory: {e}', type='negative')
                    return

                # refresh list to show contents of new directory
                self.update_list()
                dlg.close()

            with ui.row().classes('justify-end gap-2 q-mt-md'):
                ui.button('Create', on_click=_on_create)
                ui.button('Cancel', on_click=dlg.close)

        dlg.open()

    def _set_show_files(self, value: bool):
        self.show_files = bool(value)
        self.update_list()

    def _select_all(self):
        for p in self.items:
            self.selected.add(str(p))
        self._refresh_UI()

    def _clear_selection(self):
        self.selected.clear()
        self._refresh_UI()

    def _checkbox_toggled(self, index: int, checked: bool):
        """Handle checkbox toggles. Support shift-range selection when appropriate."""
        if self.multiple and self.shift_is_held and self.last_clicked_index is not None:
            start, end = sorted([self.last_clicked_index, index])
            if checked:
                for i in range(start, end + 1):
                    self.selected.add(str(self.items[i]))
            else:
                for i in range(start, end + 1):
                    self.selected.discard(str(self.items[i]))
        else:
            p_str = str(self.items[index])
            if checked:
                self.selected.add(p_str)
            else:
                self.selected.discard(p_str)
        self.last_clicked_index = index
        self._refresh_UI()

    def _label_clicked(self, index: int):
        """Clicking the label opens folders; files are not navigated by label click.

        Selection is done via the checkbox, not by clicking the label.
        """
        path = self.items[index]
        if path.is_dir():
            self.path = path
            self.path_input.value = str(self.path)
            self.last_clicked_index = None
            self.update_list()

    def update_list(self):
        # Non-blocking update: show loading indicator and spawn async worker
        self.list_container.clear()
        with self.list_container:
            with ui.row().classes('items-center justify-center q-pa-md'):
                # spinner: shows while items are being collected
                try:
                    ui.spinner()
                except Exception:
                    # fallback: simple label if spinner not available
                    ui.label('Loading...')

        # schedule the async update (collect files in a thread, then rebuild UI)
        try:
            asyncio.create_task(self._update_list_async())
        except RuntimeError:
            # no running event loop — fallback to synchronous update
            self._update_list_sync()
        
    def _refresh_UI(self):
        self.list_container.clear()
        with self.list_container:
            # Go Up item
            with ui.item(on_click=lambda: self._item_clicked(self.path.parent)).props('clickable'):
                with ui.item_section().props('avatar'):
                    ui.icon('arrow_upward')
                ui.label('..')

            for index, p in enumerate(self.items):
                p_str = str(p)
                is_selected = p_str in self.selected

                # Build item row: checkbox side for selection, icon, and clickable label for folders
                with ui.item().classes('bg-blue-100' if is_selected else ''):
                    # side: selection checkbox (separate selection control)
                    with ui.item_section().props('side'):
                        ui.checkbox(value=is_selected, on_change=lambda e, i=index: self._checkbox_toggled(i, e.value))

                    # avatar: folder/file icon
                    with ui.item_section().props('avatar'):
                        ui.icon('folder' if p.is_dir() else 'insert_drive_file',
                                color='orange' if p.is_dir() else 'blue-grey')

                    # main label: clicking it opens folders; files do not navigate
                    with ui.item_section():
                        # use a flat button for the label so we can attach a click handler
                        ui.button(p.name, on_click=lambda e, i=index: self._label_clicked(i)).props('flat').classes('text-sm text-left')

    def _collect_raw_items(self) -> List[Path]:
        raw_items: List[Path] = []
        for item in self.path.iterdir():
            try:
                item.is_dir()  # attempt to access to trigger PermissionError if any
                raw_items.append(item)
            except PermissionError:
                print("PermissionError accessing item:", item)
                # skip this item
        return raw_items
    
    def _passes_filter(self, p: Path) -> bool:
            if p.is_dir():
                return True
            elif p.is_file():
                if not self.current_filter:
                    return True if self.show_files else False
                name = p.name.lower()
                if self.regex_filter.match(name):
                    return True
            return False

    def _update_list_sync(self):
        # Synchronous fallback (same logic as the async variant)
        raw_items = self._collect_raw_items()
        self.items = [p for p in sorted(raw_items, key=lambda p: (not p.is_dir(), p.name.lower())) if self._passes_filter(p)]
        self._refresh_UI()

    async def _update_list_async(self):
        # allow the loading UI to render
        await asyncio.sleep(0)
        # collect items in a thread to avoid blocking the event loop
        try:
            raw_items = await asyncio.to_thread(self._collect_raw_items)
        except Exception as e:
            print('Error collecting items:', e)
            raw_items = []

        self.items = [p for p in sorted(raw_items, key=lambda p: (not p.is_dir(), p.name.lower())) if self._passes_filter(p)]
        self._refresh_UI()

    def _item_clicked(self, path: Path):
        self.path = path
        self.path_input.value = str(self.path)
        self.last_clicked_index = None
        self.update_list()

    def _handle_confirm(self):
        if not self.multiple and len(self.selected) > 1:
            # In single-selection mode, only keep the first selected item
            ui.notification('Please select only one item', color='red')
            return
        if self.on_select:
            self.on_select(list(self.selected))
        self.close()
    
    def _set_path(self, value: str):
        """Set path from user-edited input and refresh list."""
        p = Path(value).expanduser()
        if p.exists() and p.is_dir():
            self.path = p
            self.last_clicked_index = None
            self.update_list()
        else:
            self.list_container.clear()
            with self.list_container:
                ui.label('Invalid path').classes('text-red')
                


"""Configuration management for Spectronaut UCloud GUI."""
import json
import logging
from os import environ
from pathlib import Path
from typing import Dict, Any, Optional

DEFAULT_CONFIG = {
    'spectronaut_command': ['dotnet', '/usr/lib/spectronaut/SpectronautCMD.dll'],
    'default_dir': '/work',
    'spectronaut_key': None,  # Will be read from environment if not in config
    'port': 8080,
}


def load_config(config_path: Optional[Path|str] = None) -> Dict[str, Any]:
    """Load configuration from file or use defaults.
    
    Args:
        config_path: Path to configuration JSON file. If None, searches for
                    'config.json' in ~/.spectronaut_webui/.
    
    Returns:
        Dictionary with configuration values.
    """
    config = DEFAULT_CONFIG.copy()
    
    # Try to find config file
    if config_path is None:
        # Look in user's home directory
        config_path = Path.home().joinpath('.spectronaut_webui', 'config.json')
    else:
        config_path = Path(config_path)
    
    # Load from file if it exists
    if config_path.exists():
        try:
            with open(config_path, 'r') as f:
                user_config = json.load(f)
                config.update(user_config)
        except (json.JSONDecodeError, OSError) as e:
            # If config file is malformed or unreadable, log warning and use defaults
            logging.getLogger().warning(f"Warning: Could not load config from {config_path}: {e}")
            logging.getLogger().info("Using default configuration.")
    
    # Environment variable takes precedence over config file for license key
    env_key = environ.get('SPECTRONAUTKEY', None)
    if env_key is not None:
        config['spectronaut_key'] = env_key
    
    return config


def create_default_config(config_path: Optional[Path|str] = None) -> Path:
    """Create a default configuration file.
    
    Args:
        config_path: Path where to create the config file. If None, creates
                    'config.json' in ~/.spectronaut_webui/.
    
    Returns:
        Path to the created configuration file.
    """
    if config_path is None:
        config_dir = Path.home().joinpath('.spectronaut_webui')
        config_dir.mkdir(parents=True, exist_ok=True)
        config_path = config_dir.joinpath('config.json')
    else:
        config_path = Path(config_path)
        config_path.parent.mkdir(parents=True, exist_ok=True)
    
    with open(config_path, 'w') as f:
        json.dump(DEFAULT_CONFIG, f, indent=2)
    
    return config_path


def generate_config_cli() -> None:
    """CLI entry point to generate a default configuration file.
    
    Creates a default config.json in ~/.spectronaut_webui/ if it doesn't exist,
    or confirms the existing location.
    """
    config_path = Path.home().joinpath('.spectronaut_webui', 'config.json')
    
    if config_path.exists():
        print(f"Configuration file already exists at: {config_path}")
    else:
        created_path = create_default_config()
        print(f"Default configuration file created at: {created_path}")
        print("\nPlease edit this file to configure:")
        print("  - spectronaut_command: Command to run Spectronaut")
        print("  - spectronaut_key: License key (or set SPECTRONAUTKEY environment variable)")
        print("  - default_dir: Default working directory")
        print("  - port: Web UI port (default: 8080)")



import asyncio
import concurrent.futures
import functools
import logging
import multiprocessing
import pandas as pd
import threading
import zipfile
from typing import List, Optional, Tuple, Sequence
from pathlib import Path
from asyncio.subprocess import STDOUT, PIPE

# Global registry of active subprocesses for cancellation cleanup
_active_processes = {}
_process_lock = threading.Lock()

def track_subprocess_cleanup(coro_func):
    """Decorator that tracks running subprocesses and kills them on coroutine cancellation.
    
    When the decorated coroutine is cancelled (CancelledError), all spawned subprocesses
    are terminated gracefully first, then killed if necessary.
    
    Usage:
        @track_subprocess_cleanup
        async def my_long_running_function():
            ...
    """
    @functools.wraps(coro_func)
    async def wrapper(*args, **kwargs):
        # Get or create a unique ID for this invocation
        task_id = id(asyncio.current_task())
        _active_processes[task_id] = []
        
        try:
            result = await coro_func(*args, **kwargs)
            return result
        except asyncio.CancelledError:
            # On cancellation, clean up all subprocesses
            await _cleanup_processes(task_id)
            raise
        finally:
            # Clean up the registry
            with _process_lock:
                _active_processes.pop(task_id, None)
    
    return wrapper

async def _cleanup_processes(task_id: int):
    """Terminate all processes associated with a task."""
    with _process_lock:
        processes = _active_processes.get(task_id, [])
    
    if not processes:
        return
    
    logging.getLogger().warning(f'Cleaning up {len(processes)} subprocess(es)')
    
    # First, try to terminate gracefully
    for proc in processes:
        if proc and proc.returncode is None:
            try:
                proc.terminate()
            except Exception:
                pass
    
    # Give them a moment to terminate
    await asyncio.sleep(0.5)
    
    # Then force-kill any that are still running
    for proc in processes:
        if proc and proc.returncode is None:
            try:
                proc.kill()
            except Exception:
                pass
    
    # Wait for all processes to finish
    for proc in processes:
        if proc:
            try:
                await proc.wait() if hasattr(proc, 'wait') else None
            except Exception:
                pass

def register_subprocess(process):
    """Register a subprocess with the current task for cleanup on cancellation."""
    task = asyncio.current_task()
    if task is None:
        return
    
    task_id = id(task)
    with _process_lock:
        if task_id in _active_processes:
            _active_processes[task_id].append(process)

def _extract_zip_worker(idx: int, zip_path: Path, extract_path: Path, progress_queue=None):
    """Worker that extracts zip_path into extract_base/name and verifies analysis.tdf exists.

    Returns (idx, success:bool, real_path_str or '', error_message)
    """
    def _try_put(*args):
        if progress_queue is not None:
            try:
                progress_queue.put(*args)
            except Exception:
                pass

    try:
        _try_put(('progress', f'Extracting {str(zip_path)}'))

        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            zip_ref.extractall(extract_path)
        real_path = extract_path.joinpath('analysis.tdf')
        if not real_path.exists():
            _try_put(('done', idx, False, '', f'Missing analysis.tdf after extraction for {str(zip_path)}'))

            return (idx, False, '', f'Missing analysis.tdf after extraction for {str(zip_path)}')
        
        _try_put(('done', idx, True, str(real_path.parent), ''))
        
        return (idx, True, str(real_path.parent), '')
    
    except Exception as e:
        _try_put(('done', idx, False, '', str(e)))
        return (idx, False, '', str(e))

def validate_filetable(filetable: List[dict], rules: str ='raw') -> bool:
    types = {file['type'] for file in filetable}
    
    if rules == 'raw':
        if len(types) == 1 and 'Thermo Raw' in types:
            return True
        elif len(types) == 1 and ('Bruker D' in types or 'Bruker D Zip' in types):
            return True
        elif len(types) == 2 and 'Bruker D' in types and 'Bruker D Zip' in types:
            return True
        else:
            return False
    elif rules == 'sne':
        if len(types) == 1 and 'SNE' in types:
            return True
        else:
            return False
    else:
        return False
    
def write_conditon_file(filetable: List[dict], output_path: str, log: logging.Logger) -> None:
    df = pd.DataFrame(filetable)
    df['#'] = range(1, len(df) + 1)
    df['File Name'] = df['name'].apply(lambda x: Path(x).stem)
    df['fraction'] = df['fraction'].apply(lambda s: 'NA' if s == '' else s)
    df['condition'] = df['condition'].apply(lambda s: 'NA' if s == '' else s)
    #unfilled replicate column, fill like 1, 2, 3, ... for the same condition and fraction
    
    if df['replicate'].astype(str).apply(lambda s: s.strip().lower() in ['', 'none', 'nan']).all():
        df['replicate'] = df.groupby(['fraction', 'condition'])['replicate'].\
            transform(lambda f: range(1, f.shape[0] + 1))
    
    #semifilled replicate column
    if df['replicate'].astype(str).apply(lambda s: s.strip().lower() in ['', 'none', 'nan']).any():
        log.warning('Only some replicates were assigned, check your input')
    
    df.rename(columns={'reference': 'Reference',
                       'name': 'Run Label',
                       'condition': 'Condition',
                       'fraction': 'Fraction',
                       'replicate': 'Replicate'}, inplace=True)
    df['Label'] = df['Condition']
    df.loc[:, ['#', 'Reference', 'Run Label', 'Condition', 'Fraction',
                'Replicate', 'Label', 'File Name']].to_csv(output_path, sep='\t', index=False)

def prepare_datafiles(filetable: List[dict], data_folder: Path, log: logging.Logger, progress, cancel_event: Optional[threading.Event] = None) -> None:
    """Prepare data files for processing.
    
    Args:
        cancel_event: optional threading.Event to signal cancellation from the main task
    """
    # First, handle Bruker D folders synchronously (quick checks)
    zip_tasks: List[Tuple[int, dict]] = []  # list of (index, filedict) for zip entries
    for idx, file in enumerate(filetable):
        if file['type'] == 'Bruker D':
            real_path = Path(file['path']).joinpath('analysis.tdf')
            if not real_path.exists():
                raise FileNotFoundError(f'Corrupted Bruker D folder: {file["path"]} (missing analysis.tdf)')
            else:
                file['path'] = str(real_path.parent)

        elif file['type'] == 'Bruker D Zip':
            # defer unzipping to parallel workers
            zip_tasks.append((idx, file))
        else:
            # For Thermo Raw files, no action needed
            continue

    if zip_tasks:
        # choose number of workers (max 8 for IO bound tasks)
        max_workers = min(len(zip_tasks), 8)
        log.info(f'Extracting {len(zip_tasks)} Bruker D Zip file(s)')
        log.debug(f'Using {max_workers} workers')

        # Use a Manager queue for streaming progress from workers
        manager = multiprocessing.Manager()
        progress_q = manager.Queue()

        total = len(zip_tasks)
        done_counter = [0]
        errors = []
        progress.visible = True
        progress.value = 0
        
        # Flag to signal early exit on cancellation
        should_exit = [False]

        def _progress_reader():
            # consume progress messages and update filetable/log
            while done_counter[0] < total and not should_exit[0]:
                try:
                    item = progress_q.get(timeout=1)
                except Exception:
                    continue
                if not item:
                    continue
                typ = item[0]
                if typ == 'progress':
                    _, msg = item
                    log.info(msg)
                elif typ == 'done':
                    _, idx, success, real_path, err = item
                    file = filetable[idx]
                    if success:
                        file['path'] = real_path
                        file['name'] = Path(file['name']).stem
                        log.info(f'Extracted {file["name"]} -> {real_path}')
                    else:
                        log.error(f'Failed extracting {file["path"]}: {err}')
                        errors.append((file['path'], err))
                    done_counter[0] += 1
                    progress.value = done_counter[0]/total

        reader_thread = threading.Thread(target=_progress_reader, daemon=True)
        reader_thread.start()

        # Try process pool first, fallback to thread pool
        futures = []
        executor = None
        try:
            executor = concurrent.futures.ProcessPoolExecutor(max_workers=max_workers)
            futures = [executor.submit(_extract_zip_worker, idx, Path(file['path']),
                           Path(data_folder).joinpath(Path(file['name']).stem), progress_q) for idx, file in zip_tasks]
            # Wait for all futures to complete with periodic checks for cancellation
            # Use a timeout loop so we can check cancel_event periodically
            while True:
                done, pending = concurrent.futures.wait(futures, timeout=0.2)
                if not pending:  # All futures completed
                    break
                if cancel_event and cancel_event.is_set():
                    # Cancellation requested: cancel all pending futures
                    log.debug('Cancellation signal received, terminating extraction tasks')
                    for fut in futures:
                        fut.cancel()
                    raise asyncio.CancelledError('Zip extraction cancelled')
        except asyncio.CancelledError:
            # On cancellation, clean up and re-raise
            log.debug('Extraction cancelled, cleaning up executor')
            if executor:
                for fut in futures:
                    fut.cancel()
            raise
        except Exception as e:
            log.debug(f'Process pool extraction failed ({e})')
        finally:
            # Mark that we want to exit, but keep the queue alive for in-flight processes
            should_exit[0] = True
            
            # Cancel only futures that haven't started yet
            # This allows running processes to finish naturally
            if futures:
                for fut in futures:
                    fut.cancel()
            
            # Wait for all futures to complete (both cancelled and running)
            # This ensures all processes finish their work and close queue connections
            if futures:
                try:
                    concurrent.futures.wait(futures, timeout=5)
                except Exception as e:
                    log.debug(f'Error waiting for futures: {e}')
            
            # Now shutdown the executor after all processes have finished
            if executor:
                executor.shutdown(wait=True)
            
            # Give reader thread time to consume remaining messages
            reader_thread.join(timeout=2)
            
            # NOW all processes are dead; safe to shutdown the manager
            try:
                manager.shutdown()
            except Exception as e:
                log.debug(f'Error shutting down manager: {e}')

            progress.visible = False

        if errors:
            raise Exception('Error extracting zip file(s)')

async def prepare_datafiles_async(filetable: List[dict], data_folder: Path, log: logging.Logger, progress) -> None:
    """Async wrapper for prepare_datafiles that allows cancellation.
    
    Runs the synchronous prepare_datafiles in a thread pool so it can be cancelled.
    When cancelled, sets the cancel_event which signals the extraction threads to stop.
    """
    cancel_event = threading.Event()

    try:
        await asyncio.to_thread(prepare_datafiles, filetable, data_folder, log, progress, cancel_event)
    except asyncio.CancelledError:
        log.warning('Zip extraction cancelled by user')
        # Signal the worker thread to stop processing
        cancel_event.set()
        raise
        
def get_args(args: dict) -> Sequence[str]:
    return functools.reduce(lambda i,j: i + j, _parse_args(args), [])

def get_global_args(args: dict) -> Sequence[str]:
    result = []
    if args.get('temp_directory'):
        result.extend(['-setTemp', f'{args["temp_directory"]}'])
        args.pop('temp_directory')
    if args.get('mod_repository'):
        result.extend(['--importModRepository', f'{args["mod_repository"]}'])
        args.pop('mod_repository')
    if args.get('enzyme_database'):
        result.extend(['--importEnzymeDB', f'{args["enzyme_database"]}'])
        args.pop('enzyme_database')
    
    return result

def _parse_args(args: dict) -> Sequence[list]:
    result = []
    if args.get('protocol'):
        result.append([f'{args["protocol"]}'])
    if args.get('experiment_name'):
        result.append(['-n', f'{args["experiment_name"]}'])
    if args.get('condition_file'):
        result.append(['-con', f'{args["condition_file"]}'])
    if args.get('properties_file'):
        result.append(['-s', f'{args["properties_file"]}'])
    if args.get('report_file'):
        result.append(['-rs', f'{args["report_file"]}'])
    if args.get('fasta_file'):
        result.append(['-fasta', f'{args["fasta_file"]}'])
    if args.get('go_file'):
        result.append(['-go', f'{args["go_file"]}'])
    if args.get('output_directory'):
        result.append(['-o', f'{args["output_directory"]}'])
    if args.get('verbose'):
        result.append(['--verbose'])
    if args.get('parquet'):
        result.append(['--writeParquet'])
    if args.get('error_term'):
        result.append(['--terminateAfterError'])
    if args.get('segmented'):
        result.append(['-segmented'])
    
    return result

def get_full_args(args: dict, file_arg='-r') -> List[str]:
    args_list = functools.reduce(lambda i,j: i + j, _parse_args(args), [])
    for file in args.get('datafiles', []):
        args_list.extend([file_arg, f'{file["path"]}'])

    return args_list

async def run_cmd(args: Sequence[str], log: logging.Logger, timeout: float | None = None) -> bool:
    """Run a command asynchronously, streaming stdout/stderr to the logger.

    - args: sequence of command + arguments (same as subprocess.exec args)
    - timeout: optional timeout in seconds for the whole process
    Returns True on exit code 0, False otherwise.
    This implementation handles decode errors, supports cancellation and timeouts,
    and ensures the subprocess is terminated if the coroutine is cancelled.
    """
    log.debug(f'run_cmd {args} (timeout={timeout})')
    try:
        process = await asyncio.create_subprocess_exec(*args, stdout=PIPE, stderr=STDOUT)
        # Register this process for cleanup if the task is cancelled
        register_subprocess(process)
        log.debug(f'Added PID: {process.pid}')

        try:
            # Stream output line-by-line using async for when available
            async def _stream_stdout():
                if process.stdout is None:
                    return
                async for raw_line in process.stdout:
                    try:
                        line = raw_line.decode(errors='replace').rstrip()
                    except Exception:
                        # Safe fallback
                        line = str(raw_line)
                    log.info(line)

            if timeout is not None:
                # run streaming and wait with timeout
                await asyncio.wait_for(_stream_stdout(), timeout=timeout)
                rc = await process.wait()
            else:
                # no timeout: stream until EOF then wait for exit
                await _stream_stdout()
                rc = await process.wait()

            return rc == 0

        except asyncio.TimeoutError:
            log.error(f'Command timed out after {timeout} seconds: {args}')
            try:
                process.kill()
            except Exception:
                pass
            await process.wait()
            return False

        except asyncio.CancelledError:
            # Caller cancelled the coroutine: terminate the subprocess
            try:
                log.debug(f'Killing {process.pid}')
                process.kill()
            except Exception as e:
                log.debug(f'Killing failed for {process.pid}: {e}')
                pass
            await process.wait()
            raise

    except Exception as e:
        log.error(f'Cannot execute command {args}: {e}')
        return False



import asyncio
import logging
from pathlib import Path
from tempfile import mkdtemp
from shutil import copy
from typing import Any, Dict, List
from nicegui import app, ui
from starlette.formparsers import MultiPartParser
from .config import load_config
from .widgets import LocalPicker
from .workflows import process_direct, process_convert, process_combine

# large file upload support
MultiPartParser.spool_max_size = 1024 * 1024 * 100  # 100 MB

# --- Global constants ---
global_temp_directory = mkdtemp()
_config = None

#logging setup
class LogElementHandler(logging.Handler):
    """A logging handler that emits messages to a log element."""

    level_format = {
        logging.DEBUG: 'text-grey',
        logging.INFO: 'text-green',
        logging.WARNING: 'text-orange',
        logging.ERROR: 'text-red',
        logging.CRITICAL: 'text-red font-bold'
    }

    def __init__(self, element: ui.log, level: int = logging.NOTSET) -> None:
        self.element = element
        super().__init__(level)

    def emit(self, record: logging.LogRecord) -> None:
        try:
            msg = self.format(record)
            splits = [msg[i:i+160] for i in range(0, len(msg), 160)]
            self.element.push(f'{"\n".join(splits)}', classes=self.level_format.get(record.levelno, ''))
            # after pushing a message, instruct the browser to scroll the log area to the bottom
            try:
                #some JS madness
                ui.run_javascript("(function(){el = document.getElementById('terminal_log');\
                                  if(!el) return;\
                                  sc = el.querySelector('.q-scrollarea__content') || el.querySelector('pre') || el;\
                                  sc.scrollTop = sc.scrollHeight;})()")
            except Exception:
                # ignore JS invocation errors
                pass
        except Exception:
            self.handleError(record)

log = logging.getLogger()
logging.basicConfig(format='%(asctime)s [%(name)s] %(levelname)s: %(message)s', level=logging.DEBUG,
                    datefmt='%Y-%m-%d %H:%M:%S')

# Load configuration from file or use defaults
_config = load_config()
SPECTRONAUT = _config['spectronaut_command']
DEFAULT_DIR = _config['default_dir']
SPECTRONAUT_KEY = _config['spectronaut_key']
PORT = _config['port']

def check_type(path: Path) -> str:
    if path.is_dir() and path.suffix.lower() == '.d':
        return 'Bruker D'
    elif path.is_dir():
        return 'Folder'
    elif path.is_file() and path.suffix.lower() == '.raw':
        return 'Thermo Raw'
    elif path.is_file() and ''.join(path.suffixes[-2:]).lower() == '.d.zip':
        return 'Bruker D Zip'
    elif path.is_file() and path.suffix.lower() == '.sne':
        return 'SNE'
    else:
        return 'File'

def add_to_datafiles(paths: List[str], datafiles, table):
    added = 0
    if len(paths) > 0:
        paths.sort(key=lambda r: Path(r).name or r)

    for p in paths:
        if any(d['path'] == p for d in datafiles):
            continue
        path_obj = Path(p)
        datafiles.append({
            'name': path_obj.name or str(path_obj),
            'type': check_type(path_obj),
            'path': str(path_obj),
            'replicate': '',
            'condition': '',
            'fraction': '',
            'reference': False,
        })
        added += 1
    
    try:
        table.options['rowData'] = datafiles
    except Exception:
        pass
    table.update()
    ui.notify(f'Added {added} items', type='positive')

def open_file_picker(field, extension: str = ''):
    def _on_select(paths: List[str]):
        if not paths:
            ui.notify('Nothing selected', type='negative')
            return
        path_obj = Path(paths[0])
        if not path_obj.is_file():
                ui.notify('Selected path is not a file', type='negative')
                return
        field.value = str(path_obj)
    
    picker = LocalPicker(directory=DEFAULT_DIR, multiple=False, show_files=True,
                            default_selection=extension, on_select=_on_select)
    try:
        picker.open()
    except Exception:
        pass

def open_dirw_picker(field):
    def _on_select(paths: List[str]):
        if not paths:
            ui.notify('Nothing selected', type='negative')
            return
        try:
            path_obj = Path(paths[0])
            if not path_obj.is_dir():
                ui.notify('Selected path is not a directory', type='negative')
                return
            path_obj.joinpath('test_permission.tmp').touch()
            path_obj.joinpath('test_permission.tmp').unlink()
        except PermissionError:
            ui.notify('Permission error accessing selected path', type='negative')
            return
        field.value = paths[0]

    picker = LocalPicker(directory=DEFAULT_DIR, multiple=False, show_files=False,
                            on_select=_on_select)
    try:
        picker.open()
    except Exception:
        pass

def handle_upload(field):
    """Open a dialog with an upload control, save the uploaded file to a temp
    directory and set the input `field.value` to the saved path.
    """
    dlg = ui.dialog()
    with dlg:
        async def _on_upload(e):
            file = getattr(e, 'file', None)
            if not file:
                ui.notify('No file received', type='negative')
                return
            filename = getattr(file, 'name', None)
            if not filename:
                ui.notify('Cannot determine filename', type='negative')
                return
            
            save_path = Path(global_temp_directory).joinpath(filename)
            try:
                await file.save(save_path)
            except Exception as exc:
                ui.notify(f'Error saving uploaded file: {exc}', type='negative')
                return
            field.value = str(save_path)
            dlg.close()

        ui.upload(label='Select a file', auto_upload=True, max_files=1, on_upload=_on_upload)

    dlg.open()

# --- UI LAYER ---
columnDefs = [
    {'headerName': 'Name', 'field': 'name', 'align': 'left', 'filter': 'agTextColumnFilter'},
    {'headerName': 'Type', 'field': 'type', 'align': 'left', 'filter': 'agTextColumnFilter'},
    {'headerName': 'Absolute Path', 'field': 'path', 'align': 'left', 'hide': True},
    {'headerName': 'Reference', 'field': 'reference', 'align': 'center', 'editable': True,
     'cellRenderer': 'agCheckboxCellRenderer'},
    {'headerName': 'Condition', 'field': 'condition', 'align': 'center', 'editable': True},
    {'headerName': 'Replicate', 'field': 'replicate', 'align': 'center', 'editable': True},
    {'headerName': 'Fraction', 'field': 'fraction', 'align': 'center', 'editable': True},
]

#root element
def root():
    button_style = 'px-4 py-2 rounded-md bg-blue-500 text-white hover:bg-blue-700 transition-colors text-lg font-semibold'
    with ui.row().classes('w-full q-pa-md bg-blue-100 gap-4 items-center'):
        ui.label('Spectronaut UCloud GUI').classes('text-xl font-bold')
        with ui.row().classes('gap-2'):
            ui.link('Info', '/').classes(button_style)
            ui.link('Convert', '/convert').classes(button_style)
            ui.link('Combine', '/combine').classes(button_style)
            ui.link('DirectDIA', '/direct').classes(button_style)
            ui.link('DIA', '/dia').classes(button_style)
        ui.space()
        ui.button('Close', icon='power_settings_new', color='red', on_click=lambda: app.shutdown())

    ui.sub_pages({'/': info_page,
                  '/convert': convert_page,
                  '/combine': combine_page,
                  '/direct': directdia_page,
                  '/dia': dia_page}).classes('w-full')

def info_page():
    """Info page - starting page."""
    ui.markdown('''
    ## Spectronaut UCloud GUI

    This application helps you to setup the search using Spectronaut in UCloud 
    environment.
                
    Use the navigation bar at the top to switch between different functionalities.
                
    Each page has parameter tabs where you can select processing parameters, input files,
    and their metadata.
    
    Click "Start Processing" on the  the selected items.

    Processing log will be displayed in the Output tab.
    ''').classes('q-pa-md')

def convert_page():
    """Convert workflow page."""

    datafiles: List[dict] = []
    table = None
    running_task: Dict[str, Any] = {'task': None}

    with ui.tabs().classes('w-full') as tabs:
        param_tab = ui.tab('Parameters')
        output_tab = ui.tab('Output')

    with ui.tab_panels(tabs, value=param_tab).classes('w-full'):
        with ui.tab_panel(param_tab):
            def open_D_picker():
                picker = LocalPicker(directory=DEFAULT_DIR, multiple=True, show_files=False,
                                    on_select=lambda paths: add_to_datafiles(paths, datafiles, table))
                try:
                    picker.open()
                except Exception:
                    pass

            def open_raw_picker():
                picker = LocalPicker(directory=DEFAULT_DIR, multiple=True, show_files=True,
                                    default_selection='.raw', on_select=lambda paths: add_to_datafiles(paths, datafiles, table))
                try:
                    picker.open()
                except Exception:
                    pass
            
            def open_Dzip_picker():
                picker = LocalPicker(directory=DEFAULT_DIR, multiple=True, show_files=True,
                                    default_selection='.d.zip', on_select=lambda paths: add_to_datafiles(paths, datafiles, table))
                try:
                    picker.open()
                except Exception:
                    pass

            async def _delete_selected(*_):
                """Delete selected rows from the datafiles list and update the table."""
                rows = None
                try:
                    rows = await table.get_selected_rows() # type: ignore
                except Exception:
                    rows = None

                targets = set()
                if rows:
                    for r in rows:
                        if isinstance(r, dict) and r.get('path'):
                            targets.add(r.get('path'))

                if not targets:
                    ui.notify('No rows selected', type='negative')
                    return

                initial_count = len(datafiles)
                datafiles[:] = [d for d in datafiles if d.get('path') not in targets]

                try:
                    table.options['rowData'] = datafiles # type: ignore
                except Exception:
                    pass
                table.update() # type: ignore
                deleted_count = initial_count - len(datafiles)
                ui.notify(f'Deleted {deleted_count} items', type='positive')
            
            async def _clear_table(*_):
                """Clear all entries from the datafiles list and update the table."""
                if not datafiles:
                    ui.notify('Data table is already empty', type='warning')
                    return
                datafiles.clear()
                try:
                    table.options['rowData'] = datafiles # type: ignore
                except Exception:
                    pass
                table.update() # type: ignore
                ui.notify('Cleared all items from the data table', type='positive')

            columnDefsNew = [e.copy() for e in columnDefs]
            for c_def in columnDefsNew[3:]:
                c_def['hide'] = True
            
            with ui.row().classes('q-pa-md gap-4'):
                ui.button('Add Bruker D', icon='add', on_click=open_D_picker)
                ui.button('Add Thermo Raw', icon='add', on_click=open_raw_picker)
                ui.button('Add zipped Bruker D', icon='add', on_click=open_Dzip_picker)

            with ui.row().classes('w-full q-pa-md gap-2'):
                table = ui.aggrid({'columnDefs': columnDefsNew, 'rowData': datafiles,
                                        'rowSelection': {'mode': 'multiRow'}}).classes('h-128')
                
                ui.button('Delete selected', on_click=_delete_selected).tooltip('Delete selected entries from the data table')
                ui.button('Clear all', on_click=_clear_table).tooltip('Clear all entries from the data table')
            
            with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                prop_input = ui.input(label='Settings file', placeholder='Enter settings file path here').classes('grow')
                ui.button('Select', icon='file_open', on_click=lambda _: open_file_picker(prop_input, '.prop'))
                ui.button('Upload', icon='file_upload', on_click=lambda _: handle_upload(prop_input))
                
            with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                output_dir = ui.input(label='Output Directory', placeholder='Enter output directory here').classes('grow')
                ui.button('Select', icon='folder_open', on_click=lambda _: open_dirw_picker(output_dir))
            
            with ui.expansion('Advanced options').classes('w-full'):
                with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                    temp_dir = ui.input(label='Temp Directory', placeholder='Enter temporary directory here or leave empty for default one').classes('grow')
                    ui.button('Select', icon='folder_open', on_click=lambda _: open_dirw_picker(temp_dir))

                with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                    mark_verbose = ui.checkbox('Verbose output')
                    mark_segment = ui.checkbox('Segmented dia-PASEF (beta)')
                    mark_error = ui.checkbox('Terminate on error')

            with ui.row().classes('q-pa-md'):        
                start_button = ui.button('Start Processing', color='primary')
                # Run the processing coroutine inside the UI slot so it can safely
                # perform UI updates (avoids "slot stack is empty" runtime error).
                async def _on_start_click(*_):
                    args = {
                        'protocol': 'convert',
                        'properties_file': prop_input.value,
                        'output_directory': output_dir.value,
                        'temp_directory': temp_dir.value,
                        'verbose': mark_verbose.value,
                        'segmented': mark_segment.value,
                        'error_term': mark_error.value,
                        'datafiles': datafiles,
                    }

                    tabs.set_value('Output')
                    start_button.disable()
                    abort_button.enable()
                    ok.visible = False
                    not_ok.visible = False

                    # Create and store the task so it can be cancelled
                    running_task['task'] = asyncio.current_task()
                    try:
                        rc = await process_convert(terminal_output, progress, args, SPECTRONAUT, SPECTRONAUT_KEY)
                        if rc:
                            ok.visible = True
                            not_ok.visible = False
                        else:
                            ok.visible = False
                            not_ok.visible = True
                    except asyncio.CancelledError:
                        log.warning('Processing aborted by user')
                        ok.visible = False
                        not_ok.visible = True
                    finally:
                        running_task['task'] = None
                        start_button.enable()
                        abort_button.disable()
                        if terminate.value:
                            app.shutdown()

                start_button.on('click', _on_start_click)

        with ui.tab_panel(output_tab):
            ui.label('Console Output').classes('text-lg font-bold q-mt-lg')
            terminal_output = ui.log().props('id="terminal_log"').classes('w-full font-mono h-128')
            progress = ui.linear_progress(show_value=False).classes('w-full')
            progress.visible = False
            with ui.row().classes('w-full q-pa-md gap-2 items-center') as ok:
                ui.icon('check_circle', color='green').classes('text-5xl')
                ui.label('Success').classes('text-green text-xl')
            with ui.row().classes('w-full q-pa-md gap-2 items-center') as not_ok:
                ui.icon('error', color='red').classes('text-5xl')
                ui.label('Error').classes('text-red text-xl')
            ok.visible = False
            not_ok.visible = False

            with ui.row().classes('w-full q-pa-md gap-2'):
                abort_button = ui.button('Abort', color='negative', icon='stop')
                abort_button.disable()
                terminate = ui.checkbox('Terminate the app when processing is done')
                terminate.value = False
                
                async def _on_abort_click(*_):
                    if running_task['task'] is not None:
                        running_task['task'].cancel()
                        ui.notify('Aborting...', type='warning')
                    else:
                        ui.notify('No processing running', type='info')
                
                abort_button.on('click', _on_abort_click)
                        
            console = LogElementHandler(terminal_output)
            console.setLevel(logging.INFO)
            log.addHandler(console)
            ui.context.client.on_disconnect(lambda: log.removeHandler(console))

def combine_page():
    """Combine workflow page."""
    datafiles: List[dict] = []
    table = None
    running_task: Dict[str, Any] = {'task': None}

    with ui.tabs().classes('w-full') as tabs:
        param_tab = ui.tab('Parameters')
        output_tab = ui.tab('Output')

    with ui.tab_panels(tabs, value=param_tab).classes('w-full'):
        with ui.tab_panel(param_tab):
            def open_sne_picker():
                picker = LocalPicker(directory=DEFAULT_DIR, multiple=True, show_files=True,
                                    default_selection='.sne', on_select=lambda paths: add_to_datafiles(paths, datafiles, table))
                try:
                    picker.open()
                except Exception:
                    pass
            
            async def _delete_selected(*_):
                """Delete selected rows from the datafiles list and update the table."""
                rows = None
                try:
                    rows = await table.get_selected_rows() # type: ignore
                except Exception:
                    rows = None

                targets = set()
                if rows:
                    for r in rows:
                        if isinstance(r, dict) and r.get('path'):
                            targets.add(r.get('path'))

                if not targets:
                    ui.notify('No rows selected', type='negative')
                    return

                initial_count = len(datafiles)
                datafiles[:] = [d for d in datafiles if d.get('path') not in targets]

                try:
                    table.options['rowData'] = datafiles # type: ignore
                except Exception:
                    pass
                table.update() # type: ignore
                deleted_count = initial_count - len(datafiles)
                ui.notify(f'Deleted {deleted_count} items', type='positive')
            
            async def _clear_table(*_):
                """Clear all entries from the datafiles list and update the table."""
                if not datafiles:
                    ui.notify('Data table is already empty', type='warning')
                    return
                datafiles.clear()
                try:
                    table.options['rowData'] = datafiles # type: ignore
                except Exception:
                    pass
                table.update() # type: ignore
                ui.notify('Cleared all items from the data table', type='positive')

            columnDefsNew = [e.copy() for e in columnDefs]
            for c_def in columnDefsNew[3:]:
                c_def['hide'] = True
            
            with ui.row().classes('q-pa-md gap-4'):
                ui.button('Add SNE File', icon='add', on_click=open_sne_picker)                

            with ui.row().classes('w-full q-pa-md gap-2'):
                table = ui.aggrid({'columnDefs': columnDefsNew, 'rowData': datafiles,
                                        'rowSelection': {'mode': 'multiRow'}}).classes('h-128')
                
                ui.button('Delete selected', on_click=_delete_selected).tooltip('Delete selected entries from the data table')
                ui.button('Clear all', on_click=_clear_table).tooltip('Clear all entries from the data table')
            
            with ui.row().classes('w-full q-pa-md gap-2'):
                exp_name = ui.input(label='Experiment Name', placeholder='Enter experiment name here').classes('w-full')

            with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                prop_input = ui.input(label='Settings file', placeholder='Enter settings file path here').classes('grow')
                ui.button('Select', icon='file_open', on_click=lambda _: open_file_picker(prop_input, '.prop'))
                ui.button('Upload', icon='file_upload', on_click=lambda _: handle_upload(prop_input))
            
            with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                fasta_input = ui.input(label='FASTA file', placeholder='Enter FASTA file path here').classes('grow')
                ui.button('Select', icon='file_open', on_click=lambda _: open_file_picker(fasta_input, '.fasta|.bgsfasta'))
                ui.button('Upload', icon='file_upload', on_click=lambda _: handle_upload(fasta_input))
            
            with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                rs_input = ui.input(label='Report schema', placeholder='Enter report schema path here').classes('grow')
                ui.button('Select', icon='file_open', on_click=lambda _: open_file_picker(rs_input, '.rs'))
                ui.button('Upload', icon='file_upload', on_click=lambda _: handle_upload(rs_input))
                
            with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                output_dir = ui.input(label='Output Directory', placeholder='Enter output directory here').classes('grow')
                ui.button('Select', icon='folder_open', on_click=lambda _: open_dirw_picker(output_dir))
            
            with ui.expansion('Advanced options').classes('w-full'):
                with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                    temp_dir = ui.input(label='Temp Directory', placeholder='Enter temporary directory here or leave empty for default one').classes('grow')
                    ui.button('Select', icon='folder_open', on_click=lambda _: open_dirw_picker(temp_dir))
                with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                    modrep_input = ui.input(label='Custom modification repository', placeholder='Enter file path here').classes('grow')
                    ui.button('Select', icon='file_open', on_click=lambda _: open_file_picker(modrep_input, ''))
                    ui.button('Upload', icon='file_upload', on_click=lambda _: handle_upload(modrep_input))
                with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                    enzdb_input = ui.input(label='Custom enzyme DB', placeholder='Enter file path here').classes('grow')
                    ui.button('Select', icon='file_open', on_click=lambda _: open_file_picker(enzdb_input, '.enzdb'))
                    ui.button('Upload', icon='file_upload', on_click=lambda _: handle_upload(enzdb_input))
                with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                    mark_verbose = ui.checkbox('Verbose output')
                    mark_segment = ui.checkbox('Segmented dia-PASEF (beta)')
                    mark_parquet = ui.checkbox('Parquet output')
                    mark_error = ui.checkbox('Terminate on error')
                
            with ui.row().classes('q-pa-md'):        
                start_button = ui.button('Start Processing', color='primary')
                # Run the processing coroutine inside the UI slot so it can safely
                # perform UI updates (avoids "slot stack is empty" runtime error).
                async def _on_start_click(*_):
                    args = {
                        'protocol': 'combine',
                        'experiment_name': exp_name.value,
                        'properties_file': prop_input.value,
                        'fasta_file': fasta_input.value,
                        'report_file': rs_input.value,
                        'output_directory': output_dir.value,
                        'temp_directory': temp_dir.value,
                        'mod_repository': modrep_input.value,
                        'enzyme_database': enzdb_input.value,
                        'verbose': mark_verbose.value,
                        'segmented': mark_segment.value,
                        'parquet': mark_parquet.value,
                        'error_term': mark_error.value,
                        'datafiles': datafiles,
                    }

                    tabs.set_value('Output')
                    start_button.disable()
                    abort_button.enable()
                    ok.visible = False
                    not_ok.visible = False

                    # Create and store the task so it can be cancelled
                    running_task['task'] = asyncio.current_task()
                    try:
                        rc = await process_combine(terminal_output, progress, args, SPECTRONAUT, SPECTRONAUT_KEY)
                        if rc:
                            ok.visible = True
                            not_ok.visible = False
                        else:
                            ok.visible = False
                            not_ok.visible = True
                    except asyncio.CancelledError:
                        log.warning('Processing aborted by user')
                        ok.visible = False
                        not_ok.visible = True
                    finally:
                        running_task['task'] = None
                        start_button.enable()
                        abort_button.disable()
                        if terminate.value:
                            app.shutdown()

                start_button.on('click', _on_start_click)

        with ui.tab_panel(output_tab):
            ui.label('Console Output').classes('text-lg font-bold q-mt-lg')
            terminal_output = ui.log().props('id="terminal_log"').classes('w-full font-mono h-128')
            progress = ui.linear_progress(show_value=False).classes('w-full')
            progress.visible = False
            with ui.row().classes('w-full q-pa-md gap-2 items-center') as ok:
                ui.icon('check_circle', color='green').classes('text-5xl')
                ui.label('Success').classes('text-green text-xl')
            with ui.row().classes('w-full q-pa-md gap-2 items-center') as not_ok:
                ui.icon('error', color='red').classes('text-5xl')
                ui.label('Error').classes('text-red text-xl')
            ok.visible = False
            not_ok.visible = False

            with ui.row().classes('w-full q-pa-md gap-2'):
                abort_button = ui.button('Abort', color='negative', icon='stop')
                abort_button.disable()
                terminate = ui.checkbox('Terminate the app when processing is done')
                terminate.value = False
                
                async def _on_abort_click(*_):
                    if running_task['task'] is not None:
                        running_task['task'].cancel()
                        ui.notify('Aborting...', type='warning')
                    else:
                        ui.notify('No processing running', type='info')
                
                abort_button.on('click', _on_abort_click)
                        
            console = LogElementHandler(terminal_output)
            console.setLevel(logging.INFO)
            log.addHandler(console)
            ui.context.client.on_disconnect(lambda: log.removeHandler(console))

def directdia_page():
    """DirectDIA workflow page."""
    
    datafiles: List[dict] = []
    table = None
    running_task: Dict[str, Any] = {'task': None}  # Store the currently running coroutine task

    with ui.tabs().classes('w-full') as tabs:
        param_tab = ui.tab('Parameters')
        output_tab = ui.tab('Output')

    with ui.tab_panels(tabs, value=param_tab).classes('w-full'):
        with ui.tab_panel(param_tab):
            def open_D_picker():
                picker = LocalPicker(directory=DEFAULT_DIR, multiple=True, show_files=False,
                                    on_select=lambda paths: add_to_datafiles(paths, datafiles, table))
                try:
                    picker.open()
                except Exception:
                    pass

            def open_raw_picker():
                picker = LocalPicker(directory=DEFAULT_DIR, multiple=True, show_files=True,
                                    default_selection='.raw', on_select=lambda paths: add_to_datafiles(paths, datafiles, table))
                try:
                    picker.open()
                except Exception:
                    pass
            
            def open_Dzip_picker():
                picker = LocalPicker(directory=DEFAULT_DIR, multiple=True, show_files=True,
                                    default_selection='.d.zip', on_select=lambda paths: add_to_datafiles(paths, datafiles, table))
                try:
                    picker.open()
                except Exception:
                    pass
            
            def _on_cell_value_changed(event):
                """Update internal datafiles when a cell is edited in the grid."""
                data = None
                try:
                    data = event.args['data']
                except Exception:
                    ui.notify('Error extracting data from event', type='negative')

                if not data:
                    return
                
                column = event.args['colId']
                for d in datafiles:
                    if d.get('path') == data.get('path'):
                        if column in data:
                            d[column] = data[column]
                        break

            async def _open_condition_dialog(*_):
                await _apply_value_to_selected('condition', 'Condition', 'Condition to apply')
            
            async def _open_fraction_dialog(*_):
                await _apply_value_to_selected('fraction', 'Fraction', 'Fraction to apply')
            
            async def _assign_replicates(*_):
                """Assign replicate numbers (1,2,3,...) to rows with the same condition and fraction."""
                groups = {(d['condition'], d['fraction']): 1 for d in datafiles}
                for d in datafiles:
                    entry = groups.get((d['condition'], d['fraction']), None)
                    if entry is not None:
                        d['replicate'] = str(entry)
                        groups[(d['condition'], d['fraction'])] += 1

                try:
                    table.options['rowData'] = datafiles # type: ignore
                except Exception:
                    pass
                table.update() # type: ignore
                ui.notify(f'Assigned replicates to {len(groups)} groups', type='positive')
            
            async def _apply_value_to_selected(column: str, title: str, placeholder: str):
                """Open a small dialog to enter a string and apply it to the selected rows."""
                rows = None
                try:
                    rows = await table.get_selected_rows() # type: ignore
                except Exception:
                    rows = None

                targets = set()
                if rows:
                    for r in rows:
                        if isinstance(r, dict) and r.get('path'):
                            targets.add(r.get('path'))

                if not targets:
                    ui.notify('No rows selected', type='negative')
                    return

                with ui.dialog() as dlg, ui.card().classes('q-pa-md'):
                    async def _on_confirm(*_):
                        val = input_widget.value
                        if not val:
                            ui.notify(f'Enter a {title} value first', type='warning')
                            return
                        
                        for d in datafiles:
                            if d.get('path') in targets:
                                d[column] = val

                        try:
                            table.options['rowData'] = datafiles # type: ignore
                        except Exception:
                            pass
                        table.update() # type: ignore
                        ui.notify(f'Applied {title} "{val}" to {len(targets)} rows', type='positive')
                        dlg.close()

                    with ui.row().classes('q-pa-sm gap-2'):
                        ui.label(f'Select {title} to apply').classes('text-lg')
                        input_widget = ui.input('', placeholder=placeholder).classes('w-full')
                    with ui.row().classes('justify-end gap-2 q-mt-md'):
                        ui.button('Confirm', on_click=_on_confirm)
                        ui.button('Cancel', on_click=dlg.close)

                dlg.open()

            async def _delete_selected(*_):
                """Delete selected rows from the datafiles list and update the table."""
                rows = None
                try:
                    rows = await table.get_selected_rows()  # type: ignore
                except Exception:
                    rows = None

                targets = set()
                if rows:
                    for r in rows:
                        if isinstance(r, dict) and r.get('path'):
                            targets.add(r.get('path'))

                if not targets:
                    ui.notify('No rows selected', type='negative')
                    return

                initial_count = len(datafiles)
                datafiles[:] = [d for d in datafiles if d.get('path') not in targets]

                try:
                    table.options['rowData'] = datafiles # type: ignore
                except Exception:
                    pass
                table.update() # type: ignore
                deleted_count = initial_count - len(datafiles)
                ui.notify(f'Deleted {deleted_count} items', type='positive')
            
            async def _clear_table(*_):
                """Clear all entries from the datafiles list and update the table."""
                if not datafiles:
                    ui.notify('Data table is already empty', type='warning')
                    return
                datafiles.clear()
                try:
                    table.options['rowData'] = datafiles # type: ignore
                except Exception:
                    pass
                table.update() # type: ignore
                ui.notify('Cleared all items from the data table', type='positive')

            with ui.row().classes('q-pa-md gap-4'):
                ui.button('Add Bruker D', icon='add', on_click=open_D_picker)
                ui.button('Add Thermo Raw', icon='add', on_click=open_raw_picker)
                ui.button('Add zipped Bruker D', icon='add', on_click=open_Dzip_picker)

            with ui.row().classes('w-full q-pa-md gap-2'):
                table = ui.aggrid({'columnDefs': columnDefs, 'rowData': datafiles,
                                    'rowSelection': {'mode': 'multiRow'}}).classes('h-128')

                table.on('cellValueChanged', _on_cell_value_changed)

                ui.button('Apply condition to selected', on_click=_open_condition_dialog).tooltip('Set the same condition value to all selected rows')
                ui.button('Apply fraction to selected', on_click=_open_fraction_dialog).tooltip('Set the same fraction value to all selected rows')
                ui.button('Assign replicates', on_click=_assign_replicates).tooltip('Assign consequentive replicate numbers to the rows with the same fraction and condition')
                ui.button('Delete selected', on_click=_delete_selected).tooltip('Delete selected entries from the data table')
                ui.button('Clear all', on_click=_clear_table).tooltip('Clear all entries from the data table')
                
            with ui.row().classes('w-full q-pa-md gap-2'):
                exp_name = ui.input(label='Experiment Name', placeholder='Enter experiment name here').classes('w-full')

            with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                prop_input = ui.input(label='Properties file', placeholder='Enter properties file path here').classes('grow')
                ui.button('Select', icon='file_open', on_click=lambda _: open_file_picker(prop_input, '.prop'))
                ui.button('Upload', icon='file_upload', on_click=lambda _: handle_upload(prop_input))
                
            with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                rs_input = ui.input(label='Report schema', placeholder='Enter report schema path here').classes('grow')
                ui.button('Select', icon='file_open', on_click=lambda _: open_file_picker(rs_input, '.rs'))
                ui.button('Upload', icon='file_upload', on_click=lambda _: handle_upload(rs_input))

            with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                fasta_input = ui.input(label='FASTA file', placeholder='Enter FASTA file path here').classes('grow')
                ui.button('Select', icon='file_open', on_click=lambda _: open_file_picker(fasta_input, '.fasta|.bgsfasta'))
                ui.button('Upload', icon='file_upload', on_click=lambda _: handle_upload(fasta_input))
                
            with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                go_input = ui.input(label='GO file', placeholder='Enter GO file path here').classes('grow')
                ui.button('Select', icon='file_open', on_click=lambda _: open_file_picker(go_input, '.goannotation|.goa'))
                ui.button('Upload', icon='file_upload', on_click=lambda _: handle_upload(go_input))

            with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                output_dir = ui.input(label='Output Directory', placeholder='Enter output directory here').classes('grow')
                ui.button('Select', icon='folder_open', on_click=lambda _: open_dirw_picker(output_dir))

            with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                temp_dir = ui.input(label='Temp Directory', placeholder='Enter temporary directory here or leave empty for default one').classes('grow')
                ui.button('Select', icon='folder_open', on_click=lambda _: open_dirw_picker(temp_dir))

            with ui.expansion('Advanced options').classes('w-full'):
                with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                    modrep_input = ui.input(label='Custom modification repository', placeholder='Enter file path here').classes('grow')
                    ui.button('Select', icon='file_open', on_click=lambda _: open_file_picker(modrep_input, ''))
                    ui.button('Upload', icon='file_upload', on_click=lambda _: handle_upload(modrep_input))
                with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                    enzdb_input = ui.input(label='Custom enzyme DB', placeholder='Enter file path here').classes('grow')
                    ui.button('Select', icon='file_open', on_click=lambda _: open_file_picker(enzdb_input, '.enzdb'))
                    ui.button('Upload', icon='file_upload', on_click=lambda _: handle_upload(enzdb_input))
                with ui.row().classes('w-full q-pa-md gap-2 items-center'):
                    mark_verbose = ui.checkbox('Verbose output')
                    mark_segment = ui.checkbox('Segmented dia-PASEF (beta)')
                    mark_parquet = ui.checkbox('Parquet output')
                    mark_error = ui.checkbox('Terminate on error')

            with ui.row().classes('q-pa-md'):
                start_button = ui.button('Start Processing', color='primary')

                async def _on_start_click(*_):
                    args = {
                        'protocol': 'direct',
                        'experiment_name': exp_name.value,
                        'properties_file': prop_input.value,
                        'fasta_file': fasta_input.value,
                        'go_file': go_input.value,
                        'report_file': rs_input.value,
                        'output_directory': output_dir.value,
                        'temp_directory': temp_dir.value,
                        'verbose': mark_verbose.value,
                        'parquet': mark_parquet.value,
                        'segmented': mark_segment,
                        'error_term': mark_error.value,
                        'mod_repository': modrep_input.value,
                        'enzyme_database': enzdb_input.value,
                        'datafiles': datafiles,
                    }

                    tabs.set_value('Output')
                    ok.visible = False
                    not_ok.visible = False
                    start_button.disable()
                    abort_button.enable()
                    
                    # Create and store the task so it can be cancelled
                    running_task['task'] = asyncio.current_task()
                    try:
                        rc = await process_direct(terminal_output, progress, args, SPECTRONAUT, SPECTRONAUT_KEY)
                        if rc:
                            ok.visible = True
                            not_ok.visible = False
                        else:
                            ok.visible = False
                            not_ok.visible = True
                    except asyncio.CancelledError:
                        log.warning('Processing aborted by user')
                        ok.visible = False
                        not_ok.visible = True
                    finally:
                        running_task['task'] = None
                        start_button.enable()
                        abort_button.disable()
                        if terminate.value:
                            app.shutdown()

                start_button.on('click', _on_start_click)

        with ui.tab_panel(output_tab):
            ui.label('Console Output').classes('text-lg font-bold q-mt-lg')
            terminal_output = ui.log(max_lines=256).props('id="terminal_log"').classes('w-full font-mono h-128')
            progress = ui.linear_progress(show_value=False).classes('w-full')
            progress.visible = False
            
            with ui.row().classes('w-full q-pa-md gap-2'):
                abort_button = ui.button('Abort', color='negative', icon='stop')
                abort_button.disable()
                terminate = ui.checkbox('Terminate the app when processing is done')
                terminate.value = False
                
                async def _on_abort_click(*_):
                    if running_task['task'] is not None:
                        running_task['task'].cancel()
                        ui.notify('Aborting...', type='warning')
                    else:
                        ui.notify('No processing running', type='info')
                
                abort_button.on('click', _on_abort_click)
            
            with ui.row().classes('w-full q-pa-md gap-2 items-center') as ok:
                ui.icon('check_circle', color='green').classes('text-5xl')
                ui.label('Success').classes('text-green text-xl')
            with ui.row().classes('w-full q-pa-md gap-2 items-center') as not_ok:
                ui.icon('error', color='red').classes('text-5xl')
                ui.label('Error').classes('text-red text-xl')
            ok.visible = False
            not_ok.visible = False
            
            console = LogElementHandler(terminal_output)
            console.setLevel(logging.INFO)
            log.addHandler(console)
            ui.context.client.on_disconnect(lambda: log.removeHandler(console))

def dia_page():
    """DIA workflow page."""
    ui.label('Coming soon!').classes('q-pa-md')

def main():
    """Main entry point."""
    if SPECTRONAUT_KEY is None:
        log.error('Cannot find license key')
        exit(1)

    ui.run(root, title='Spectronaut UCloud GUI', port=PORT, reload=False)

if __name__ in {"__main__", "__mp_main__"}:
    main()



"""Workflow processing functions for Spectronaut operations."""

from .direct import process_direct
from .convert import process_convert
from .combine import process_combine

__all__ = ['process_direct', 'process_convert', 'process_combine']
